use prost;

use std::fmt::{Debug, Formatter};
use std::iter::Iterator;
use std::path::Path;
use std::sync::Arc;

use tonic::{
    codegen::BoxStream, metadata::Ascii, metadata::MetadataValue, Code, Request, Response, Status,
};
use tracing::{info, instrument};

use hg::changelog::{Changelog, ChangelogEntry};
use hg::errors::HgError;
use hg::repo::Repo;
use hg::revlog::{GraphError, NodePrefix, Revision, RevlogError, NULL_REVISION};
use hg::utils::hg_path::HgPath;
use hg::AncestorsIterator;

use super::blob::stream_blob;
use crate::config::Config;
use crate::errors::{
    status_with_structured_error, FromPathError, FromResolveRevisionError, PathErrorType,
};
use crate::git::{self, GitPathSpec, GitRevSpec};
use crate::gitaly::commit_service_client::CommitServiceClient;
use crate::gitaly::commit_service_server::{CommitService, CommitServiceServer};
use crate::gitaly::{
    find_commits_request, get_tree_entries_error, get_tree_entries_request::SortBy,
    list_commits_by_ref_name_response, tree_entry::EntryType, tree_entry_response::ObjectType,
    CommitIsAncestorRequest, CommitIsAncestorResponse, CountCommitsRequest, CountCommitsResponse,
    FindCommitRequest, FindCommitResponse, FindCommitsRequest, FindCommitsResponse,
    GetTreeEntriesError, GetTreeEntriesRequest, GetTreeEntriesResponse, LastCommitForPathRequest,
    LastCommitForPathResponse, ListCommitsByOidRequest, ListCommitsByOidResponse,
    ListCommitsByRefNameRequest, ListCommitsByRefNameResponse, PaginationParameter, PathError,
    Repository, ResolveRevisionError, TreeEntry, TreeEntryRequest, TreeEntryResponse,
};
use crate::gitlab::revision::{
    blocking_gitlab_revision_node_prefix, gitlab_revision_node_prefix, RefOrRevlogError,
};
use crate::mercurial::{
    changelog_entry_manifest, ls_path, read_blob, DirIteratorWithFlatPaths,
    DirIteratorWithoutFlatPaths, LimitedAncestorsIterator, PathContent, RecursiveDirIterator,
};
use crate::message;
use crate::metadata::correlation_id;
use crate::oid::tree_oid;
use crate::repository::{
    default_repo_spec_error_status, load_changelog_and_stream, load_changelog_and_then,
    repo_store_vfs, unimplemented_if_largefiles, RequestWithRepo,
};
use crate::sidecar;
use crate::streaming::{
    empty_response_stream, stream_chunks, stream_with_pagination, BlockingResponseSender,
    PaginableMessageItem, ResultResponseStream,
};
use crate::util::{bytes_strings_as_str, tracing_span_id};

#[derive(Debug)]
pub struct CommitServiceImpl {
    config: Arc<Config>,
    sidecar_servers: Arc<sidecar::Servers>,
}

const COMMITS_CHUNK_SIZE: usize = 50;

/// Return the [`ListCommitsByOidResponse`] for a chunk of oids (changeset ids)
///
/// Behaving like Gitaly imposes that
/// - oids that cannot be found in the repository are ignored. This is actually one of the
///   use cases for the `ListCommitsByOid` call: it can be called indiscriminately with anything
///   that looks like a hash, in particular from Markdown renderings
/// - oids that are not even hexadecimal are ignored, too. This is possibly coincidental.
///
/// [`RevlogError`] is returned in cases of repository corruption.
fn commits_by_oid_chunk(
    cl: &Changelog,
    oids: &[String],
) -> Result<ListCommitsByOidResponse, RevlogError> {
    // Not using Vec::with_capacity(oids.len()), because clients will often call this with a single
    // argument, precisely to know whether it is a commit id prefix in this repo (happens a lot
    // in Markdown rendering), and this will result in needless allocations.
    let mut commits = Vec::new();
    for oid in oids {
        match NodePrefix::from_hex(oid) {
            Err(_) => {
                continue; // non-hexadecimal values for oid are just ignored by Gitaly
            }
            Ok(node_pref) => match cl.rev_from_node(node_pref) {
                Ok(NULL_REVISION) => (),
                Ok(rev) => commits.push(message::commit(cl.entry_for_rev(rev.into())?)?),
                Err(RevlogError::InvalidRevision) => (),
                Err(e) => {
                    return Err(e);
                }
            },
        }
    }
    Ok(ListCommitsByOidResponse { commits })
}

fn commits_by_ref_name_chunk(
    store_vfs: &Path,
    cl: &Changelog,
    ref_names: &[Vec<u8>],
) -> Result<ListCommitsByRefNameResponse, RefOrRevlogError> {
    let mut commit_refs = Vec::with_capacity(ref_names.len());
    for ref_name in ref_names {
        match blocking_gitlab_revision_node_prefix(store_vfs, ref_name) {
            Ok(Some(np)) => {
                let rev = cl.rev_from_node(np)?;
                commit_refs.push(list_commits_by_ref_name_response::CommitForRef {
                    commit: Some(message::commit(cl.entry_for_rev(rev.into())?)?),
                    ref_name: ref_name.clone(),
                });
            }
            Ok(None) => {}
            Err(e) => {
                return Err(e.into());
            }
        }
    }
    Ok(ListCommitsByRefNameResponse { commit_refs })
}

fn stream_find_commits_ancestors(
    cl: &Changelog,
    node_prefix: NodePrefix,
    offset: usize,
    limit: usize,
    tx: &BlockingResponseSender<FindCommitsResponse>,
) -> Result<(), RevlogError> {
    let init_rev = cl.rev_from_node(node_prefix)?;
    let iter = LimitedAncestorsIterator::new(
        AncestorsIterator::new(cl, [init_rev], NULL_REVISION, true)
            .map_err(revlog_err_from_graph_err)?,
        offset,
        limit,
    );
    stream_chunks(
        tx,
        iter.map(|r| {
            r.map_err(|e| RevlogError::Other(revlog_err_from_graph_err(e)))
                .and_then(|r| cl.entry_for_rev(r.into()))
                .and_then(|e| message::commit(e))
        }),
        |chunk, _is_first| FindCommitsResponse { commits: chunk },
        |e| {
            Status::internal(format!(
                "Repository corruption in ancestors iteration: {}",
                e
            ))
        },
    );
    Ok(())
}

fn not_found_for_path(_path: &[u8]) -> Status {
    Status::not_found("tree entry not found")
}

fn stream_tree_entry(
    req: TreeEntryRequest,
    repo: &Repo,
    changelog: &Changelog,
    node_prefix: NodePrefix,
    tx: &BlockingResponseSender<TreeEntryResponse>,
) -> Result<(), Status> {
    let manifestlog = repo
        .manifestlog()
        .map_err(|e| Status::internal(format!("Error getting manifestlog: {:?}", e)))?;
    let (cs_node, manifest) = changelog_entry_manifest(changelog, &manifestlog, node_prefix)
        .map_err(|e| Status::internal(format!("Error obtaining manifest: {:?}", e)))?
        .ok_or_else(|| not_found_for_path(&req.path))?;

    let mut pl = req.path.len();
    while pl > 0 && req.path[pl - 1] == b'/' {
        pl -= 1;
    }
    let trimmed = &req.path[0..pl];

    match ls_path(&manifest, trimmed)
        .map_err(|e| Status::internal(format!("Error reading manifest {:?}", e)))?
    {
        PathContent::File(manifest_entry) => {
            let hg_path = HgPath::new(&req.path);
            let (mut data, metadata) = read_blob(repo, &cs_node, manifest_entry, hg_path)
                .map_err(|e| Status::internal(format!("Error reading file {:?}", e)))?;
            if req.max_size != 0 && metadata.size > req.max_size {
                tx.send(Err(Status::failed_precondition(format!(
                    "object size ({}) is bigger than the maximum allowed size ({})",
                    metadata.size, req.max_size
                ))));
            } else {
                if req.limit > 0 {
                    data.truncate(req.limit as usize);
                }
                stream_blob(tx, data, metadata)
            }
        }
        PathContent::NotFound => {
            return Err(not_found_for_path(&req.path));
        }
        PathContent::Directory(listing) => {
            let mut size: usize = 0;
            // See Python impl for explanation (note that the Rust implementation is using
            // the relative path instead of the absolute path).
            for elt in listing.iter() {
                if elt.is_file() {
                    size += 28
                } else {
                    size += 27
                }
                size += elt.relative_path.len();
            }
            tx.send(Ok(TreeEntryResponse {
                r#type: ObjectType::Tree as i32,
                oid: tree_oid(&cs_node, trimmed),
                mode: git::OBJECT_MODE_TREE,
                size: size as i64,
                ..Default::default()
            }));
        }
    }
    Ok(())
}

fn stream_get_tree_entries(
    req: GetTreeEntriesRequest,
    repo: &Repo,
    changelog: &Changelog,
    node_prefix: NodePrefix,
    tx: &BlockingResponseSender<GetTreeEntriesResponse>,
) -> Result<(), Status> {
    let manifestlog = repo
        .manifestlog()
        .map_err(|e| Status::internal(format!("Error getting manifestlog: {:?}", e)))?;
    let (cs_node, manifest) = changelog_entry_manifest(changelog, &manifestlog, node_prefix)
        .map_err(|e| Status::internal(format!("Error obtaining manifest: {:?}", e)))?
        .ok_or_else(|| not_found_for_path(&req.path))?;

    let mut pl = req.path.len();
    while pl > 0 && req.path[pl - 1] == b'/' {
        pl -= 1;
    }
    if pl == 1 && req.path[0] == b'.' {
        pl = 0
    };
    let trimmed = &req.path[0..pl];

    let commit_oid = String::from_utf8_lossy(&req.revision).to_string();

    let non_empty_tree = if req.recursive {
        stream_get_tree_entries_from_iterator(
            tx,
            &req,
            RecursiveDirIterator::new(commit_oid, cs_node, manifest.iter(), trimmed),
        )
    } else if req.skip_flat_paths {
        stream_get_tree_entries_from_iterator(
            tx,
            &req,
            DirIteratorWithoutFlatPaths::new(commit_oid, cs_node, manifest.iter(), trimmed),
        )
    } else {
        stream_get_tree_entries_from_iterator(
            tx,
            &req,
            DirIteratorWithFlatPaths::new(commit_oid, cs_node, manifest.iter(), trimmed),
        )
    };
    if !non_empty_tree {
        // There is no such thing as an empty Mercurial directory
        // this is potentially long and unnecessary in almost all requests,
        // this is why we are doing it only if the iterator ends up being
        // empty.
        return match ls_path(&manifest, trimmed) {
            Ok(PathContent::File(_)) => Err(status_with_structured_error(
                Code::InvalidArgument,
                "path not treeish",
                GetTreeEntriesError::resolve_revision_error(req.revision.clone()),
            )),
            Ok(PathContent::NotFound) => {
                let code = if req.recursive {
                    Code::NotFound
                } else {
                    Code::InvalidArgument
                };
                Err(status_with_structured_error(
                    code,
                    "invalid revision or path",
                    GetTreeEntriesError::resolve_revision_error(req.revision.clone()),
                ))
            }
            Ok(PathContent::Directory(_)) => Err(Status::internal(
                "Empty GetTreeEntries iteration, but contradiction when analyzing \
                 for error return",
            )),
            Err(e) => Err(Status::internal(format!(
                "Empty GetTreeEntries iteration and further error when analyzing \
                 for error return: {:?}",
                e
            ))),
        };
    }
    Ok(())
}

/// Stream `GetTreeEntries` responses, using the inner iterator.
///
/// Various options (e.g., `recursive`) give rise to different types of iterators on entries.
/// This function applies the sort and pagination parameters, final chunking and streams responses
/// to the channel.
fn stream_get_tree_entries_from_iterator<It>(
    tx: &BlockingResponseSender<GetTreeEntriesResponse>,
    request: &GetTreeEntriesRequest,
    iterator: It,
) -> bool
where
    It: Iterator<Item = Result<TreeEntry, HgError>>,
{
    let pagination = &request.pagination_params;
    if request.sort == SortBy::TreesFirst as i32 {
        let mut trees = Vec::new();
        let mut blobs = Vec::new();
        for res in iterator {
            match res {
                Ok(ent) => {
                    if ent.r#type == EntryType::Tree as i32 {
                        trees.push(Ok(ent));
                    } else if ent.r#type == EntryType::Blob as i32 {
                        blobs.push(Ok(ent));
                    }
                }
                // arbitrary choice, weakly motivated by errors being more likely with blobs
                // than with trees at this point.
                err => blobs.push(err),
            }
        }
        stream_get_tree_entries_paginated(tx, pagination, trees.into_iter().chain(blobs))
    } else {
        stream_get_tree_entries_paginated(tx, pagination, iterator)
    }
}

/// Stream [`GetTreeEntriesResponse`] messages with pagination
///
/// This function takes care of the offset and limit logic enclosed into the given pagination
/// parameter, regrouping the `TreeEntry` elements yielded by the iterator into suitable chunks,
/// and providing the `next_cursor` value, but does not take care of ordering.
/// Instead it expects the incoming iiterator to be already properly ordered.
fn stream_get_tree_entries_paginated<It>(
    tx: &BlockingResponseSender<GetTreeEntriesResponse>,
    pagination: &Option<PaginationParameter>,
    iterator: It,
) -> bool
where
    It: Iterator<Item = Result<TreeEntry, HgError>>,
{
    stream_with_pagination(
        tx,
        pagination,
        iterator,
        |entries, pagination_cursor| GetTreeEntriesResponse {
            entries,
            pagination_cursor,
        },
        |err| Status::internal(format!("error reading manifest: #{:?}", err)),
        |token| Status::internal(format!("could not find starting OID: {}", token)),
    )
}

impl RequestWithRepo for TreeEntryRequest {
    fn repository_ref(&self) -> Option<&Repository> {
        self.repository.as_ref()
    }
}

impl RequestWithRepo for CountCommitsRequest {
    fn repository_ref(&self) -> Option<&Repository> {
        self.repository.as_ref()
    }
}

impl RequestWithRepo for GetTreeEntriesRequest {
    fn repository_ref(&self) -> Option<&Repository> {
        self.repository.as_ref()
    }
}

impl RequestWithRepo for ListCommitsByOidRequest {
    fn repository_ref(&self) -> Option<&Repository> {
        self.repository.as_ref()
    }
}

impl RequestWithRepo for ListCommitsByRefNameRequest {
    fn repository_ref(&self) -> Option<&Repository> {
        self.repository.as_ref()
    }
}

impl RequestWithRepo for LastCommitForPathRequest {
    fn repository_ref(&self) -> Option<&Repository> {
        self.repository.as_ref()
    }
}

impl RequestWithRepo for FindCommitRequest {
    fn repository_ref(&self) -> Option<&Repository> {
        self.repository.as_ref()
    }
}

impl RequestWithRepo for FindCommitsRequest {
    fn repository_ref(&self) -> Option<&Repository> {
        self.repository.as_ref()
    }
}

impl PaginableMessageItem for TreeEntry {
    fn match_token(&self, token: &str) -> bool {
        self.oid == token
    }

    fn next_cursor(&self) -> String {
        self.oid.clone()
    }
}

#[tonic::async_trait]
impl CommitService for CommitServiceImpl {
    async fn commit_is_ancestor(
        &self,
        request: Request<CommitIsAncestorRequest>,
    ) -> Result<Response<CommitIsAncestorResponse>, Status> {
        sidecar::unary!(self, request, CommitServiceClient, commit_is_ancestor)
    }

    async fn list_commits_by_oid(
        &self,
        request: Request<ListCommitsByOidRequest>,
    ) -> Result<Response<BoxStream<ListCommitsByOidResponse>>, Status> {
        let (metadata, _ext, inner) = request.into_parts();

        self.inner_list_commits_by_oid(inner, correlation_id(&metadata))
            .await
    }

    async fn list_commits_by_ref_name(
        &self,
        request: Request<ListCommitsByRefNameRequest>,
    ) -> Result<Response<BoxStream<ListCommitsByRefNameResponse>>, Status> {
        let (metadata, _ext, inner) = request.into_parts();

        self.inner_list_commits_by_ref_name(inner, correlation_id(&metadata))
            .await
    }

    async fn find_commits(
        &self,
        request: Request<FindCommitsRequest>,
    ) -> Result<Response<BoxStream<FindCommitsResponse>>, Status> {
        sidecar::fallback_server_streaming!(
            self,
            inner_find_commits,
            request,
            CommitServiceClient,
            find_commits
        )
    }

    async fn last_commit_for_path(
        &self,
        request: Request<LastCommitForPathRequest>,
    ) -> Result<Response<LastCommitForPathResponse>, Status> {
        let (metadata, _ext, inner) = request.into_parts();

        self.inner_last_commit_for_path(inner, correlation_id(&metadata))
            .await
    }

    async fn find_commit(
        &self,
        request: Request<FindCommitRequest>,
    ) -> Result<Response<FindCommitResponse>, Status> {
        let (metadata, _ext, inner) = request.into_parts();

        self.inner_find_commit(inner, correlation_id(&metadata))
            .await
    }

    async fn tree_entry(
        &self,
        request: Request<TreeEntryRequest>,
    ) -> Result<Response<BoxStream<TreeEntryResponse>>, Status> {
        sidecar::fallback_server_streaming!(
            self,
            inner_tree_entry,
            request,
            CommitServiceClient,
            tree_entry
        )
    }

    async fn count_commits(
        &self,
        request: Request<CountCommitsRequest>,
    ) -> Result<Response<CountCommitsResponse>, Status> {
        sidecar::fallback_unary!(
            self,
            inner_count_commits,
            request,
            CommitServiceClient,
            count_commits
        )
    }

    async fn get_tree_entries(
        &self,
        request: Request<GetTreeEntriesRequest>,
    ) -> Result<Response<BoxStream<GetTreeEntriesResponse>>, Status> {
        let (metadata, _ext, inner) = request.into_parts();

        self.inner_get_tree_entries(inner, correlation_id(&metadata))
            .await
    }
}

struct TreeEntryTracingRequest<'a>(&'a TreeEntryRequest);

impl Debug for TreeEntryTracingRequest<'_> {
    fn fmt(&self, f: &mut Formatter<'_>) -> std::fmt::Result {
        f.debug_struct("TreeEntryRequest")
            .field("repository", &self.0.repository)
            .field("revision", &String::from_utf8_lossy(&self.0.revision))
            .field("path", &String::from_utf8_lossy(&self.0.path))
            .field("limit", &self.0.limit)
            .field("max_size", &self.0.max_size)
            .finish()
    }
}

struct GetTreeEntriesTracingRequest<'a>(&'a GetTreeEntriesRequest);

impl Debug for GetTreeEntriesTracingRequest<'_> {
    fn fmt(&self, f: &mut Formatter<'_>) -> std::fmt::Result {
        f.debug_struct("GetTreeEntriesRequest")
            .field("repository", &self.0.repository)
            .field("revision", &String::from_utf8_lossy(&self.0.revision))
            .field("path", &String::from_utf8_lossy(&self.0.path))
            .field("recursive", &self.0.recursive)
            .field("sort", &self.0.sort)
            .field("pagination_params", &self.0.pagination_params)
            .field("skip_flat_paths", &self.0.skip_flat_paths)
            .finish()
    }
}

struct ListCommitsByOidTracingRequest<'a>(&'a ListCommitsByOidRequest);

impl Debug for ListCommitsByOidTracingRequest<'_> {
    fn fmt(&self, f: &mut Formatter<'_>) -> std::fmt::Result {
        f.debug_struct("LisCommitsByOidRequest")
            .field("repository", &self.0.repository)
            .field("oid", &self.0.oid)
            .finish()
    }
}

struct LastCommitForPathTracingRequest<'a>(&'a LastCommitForPathRequest);

impl Debug for LastCommitForPathTracingRequest<'_> {
    fn fmt(&self, f: &mut Formatter<'_>) -> std::fmt::Result {
        f.debug_struct("LastCommitForPathRequest")
            .field("repository", &self.0.repository)
            .field("revision", &String::from_utf8_lossy(&self.0.revision))
            .field("path", &String::from_utf8_lossy(&self.0.path))
            .field("literal_pathspec", &self.0.literal_pathspec)
            .finish()
    }
}

struct FindCommitTracingRequest<'a>(&'a FindCommitRequest);

impl Debug for FindCommitTracingRequest<'_> {
    fn fmt(&self, f: &mut Formatter<'_>) -> std::fmt::Result {
        f.debug_struct("FindCommitRequest")
            .field("repository", &self.0.repository)
            .field("revision", &String::from_utf8_lossy(&self.0.revision))
            .finish()
    }
}

struct ListCommitsByRefNameTracingRequest<'a>(&'a ListCommitsByRefNameRequest);

impl Debug for ListCommitsByRefNameTracingRequest<'_> {
    fn fmt(&self, f: &mut Formatter<'_>) -> std::fmt::Result {
        f.debug_struct("ListCommitsByRefNameRequest")
            .field("repository", &self.0.repository)
            .field("ref_names", &bytes_strings_as_str(&self.0.ref_names))
            .finish()
    }
}

struct FindCommitsTracingRequest<'a>(&'a FindCommitsRequest);

impl Debug for FindCommitsTracingRequest<'_> {
    fn fmt(&self, f: &mut Formatter<'_>) -> std::fmt::Result {
        f.debug_struct("FindCommitsRequest")
            .field("repository", &self.0.repository)
            .field("revision", &String::from_utf8_lossy(&self.0.revision))
            .field("follow", &self.0.follow)
            .field("paths", &bytes_strings_as_str(&self.0.paths))
            .finish()
    }
}

impl prost::Name for GetTreeEntriesError {
    const NAME: &'static str = "GetTreeEntriesError";
    const PACKAGE: &'static str = "gitaly";
}

impl FromPathError for GetTreeEntriesError {
    fn from_path_error(err: PathError) -> Self {
        GetTreeEntriesError {
            error: Some(get_tree_entries_error::Error::Path(err)),
        }
    }
}

impl FromResolveRevisionError for GetTreeEntriesError {
    fn from_resolve_revision_error(err: ResolveRevisionError) -> Self {
        GetTreeEntriesError {
            error: Some(get_tree_entries_error::Error::ResolveTree(err)),
        }
    }
}

/// in-thread implementation for `LastCommitForPath`
///
/// This means that this is synchronmous code, running in the separate thread
/// than opens repo and changelog.
///
/// This is in particular useful to concentrate `RevlogError` in one return
/// type, to avoid littering with map_err everywhere
fn lcfp_in_thread(
    req: LastCommitForPathRequest,
    cl: &Changelog,
    node_prefix: NodePrefix,
    literal_pathspec: bool,
) -> Result<LastCommitForPathResponse, RevlogError> {
    let rev = cl.rev_from_node(node_prefix)?;
    let maybe_entry = if literal_pathspec {
        lcfp_ancestors_walk(git::GitLiteralPathSpec::new(&req.path), cl, rev)
    } else {
        lcfp_ancestors_walk(git::GitWildcardPathSpec::new(&req.path), cl, rev)
    }?;

    maybe_entry.map_or_else(
        || Ok(LastCommitForPathResponse::default()), // no match
        |cl_entry| {
            Ok(LastCommitForPathResponse {
                commit: Some(message::commit(cl_entry)?),
            })
        },
    )
}

/// Convert a GraphError to RevlogError
///
/// Talk upstream about this. When both are intertwined, it is painful not to have the
/// `From<GraphError>` implementation.
fn revlog_err_from_graph_err(err: GraphError) -> HgError {
    HgError::CorruptedRepository(format!("{:?}", err))
}

fn lcfp_ancestors_walk(
    path_spec: impl GitPathSpec,
    cl: &Changelog,
    rev: Revision,
) -> Result<Option<ChangelogEntry>, RevlogError> {
    if path_spec.matches_root() {
        return Ok(Some(cl.entry_for_rev(rev.into())?));
    }
    for anc_rev in
        AncestorsIterator::new(cl, [rev], NULL_REVISION, true).map_err(revlog_err_from_graph_err)?
    {
        let cl_entry = cl.entry_for_rev(anc_rev.map_err(revlog_err_from_graph_err)?.into())?;
        let cl_data = cl_entry.data()?;
        for cl_path in cl_data.files() {
            if path_spec.matches(cl_path.as_bytes()) {
                return Ok(Some(cl_entry));
            }
        }
    }
    Ok(None)
}

impl CommitServiceImpl {
    #[instrument(name = "tree_entry", skip(self, request), fields(span_id))]
    async fn inner_tree_entry(
        &self,
        request: &TreeEntryRequest,
        correlation_id: Option<&MetadataValue<Ascii>>,
    ) -> ResultResponseStream<TreeEntryResponse> {
        tracing_span_id!();
        info!("Processing request={:?}", TreeEntryTracingRequest(request));

        if request.revision.is_empty() {
            return Err(Status::invalid_argument("empty revision"));
        }

        let config = self.config.clone();
        let store_vfs = repo_store_vfs(&config, &request.repository)
            .await
            .map_err(default_repo_spec_error_status)?;

        unimplemented_if_largefiles(&self.config, &request.repository).await?;

        match gitlab_revision_node_prefix(&store_vfs, &request.revision)
            .await
            .map_err(|e| Status::internal(format!("Error resolving revision: {:?}", e)))?
        {
            None => {
                info!("Revision not resolved");
                Err(not_found_for_path(&request.path))
            }
            Some(node_prefix) => {
                info!("Revision resolved as {:x}", &node_prefix);

                load_changelog_and_stream(
                    self.config.clone(),
                    request.clone(),
                    default_repo_spec_error_status,
                    move |req, repo, cl, tx| {
                        if let Err(e) = stream_tree_entry(req, repo, cl, node_prefix, &tx) {
                            tx.send(Err(e));
                        }
                    },
                )
            }
        }
    }

    #[instrument(name = "count_commits", skip(self, request), fields(span_id))]
    async fn inner_count_commits(
        &self,
        request: &CountCommitsRequest,
        correlation_id: Option<&MetadataValue<Ascii>>,
    ) -> Result<Response<CountCommitsResponse>, Status> {
        if request.revision.is_empty() && !request.all {
            // minimal case to prove that fallback works
            Err(Status::invalid_argument("empty Revision and false All"))
        } else {
            Err(Status::unimplemented("soon"))
        }
    }

    #[instrument(name = "get_tree_entries", skip(self, request), fields(span_id))]
    async fn inner_get_tree_entries(
        &self,
        request: GetTreeEntriesRequest,
        correlation_id: Option<&MetadataValue<Ascii>>,
    ) -> ResultResponseStream<GetTreeEntriesResponse> {
        tracing_span_id!();
        info!(
            "Processing, request={:?}",
            GetTreeEntriesTracingRequest(&request)
        );

        if request.revision.is_empty() {
            return Err(Status::invalid_argument("empty revision"));
        }

        if request.path.is_empty() {
            return Err(status_with_structured_error(
                Code::InvalidArgument,
                "empty path",
                GetTreeEntriesError::path_error(request.path, PathErrorType::EmptyPath),
            ));
        }

        let config = self.config.clone();
        let store_vfs = repo_store_vfs(&config, &request.repository)
            .await
            .map_err(default_repo_spec_error_status)?;

        match gitlab_revision_node_prefix(&store_vfs, &request.revision)
            .await
            .map_err(|e| Status::internal(format!("Error resolving revision: {:?}", e)))?
        {
            None => Err(status_with_structured_error(
                Code::InvalidArgument,
                "invalid revision or path",
                GetTreeEntriesError::resolve_revision_error(request.revision.clone()),
            )),
            Some(node_prefix) => {
                info!("Revision resolved as {:x}", &node_prefix);

                if request.pagination_params.as_ref().map(|p| p.limit) == Some(0) {
                    // let's not even load the repo (see Gitaly Comparison tests,
                    // no error case either)
                    return empty_response_stream();
                }

                load_changelog_and_stream(
                    self.config.clone(),
                    request,
                    default_repo_spec_error_status,
                    move |req, repo, cl, tx| {
                        if let Err(e) = stream_get_tree_entries(req, repo, cl, node_prefix, &tx) {
                            tx.send(Err(e));
                        }
                    },
                )
            }
        }
    }

    #[instrument(name = "list_commits_by_oid", skip(self, request), fields(span_id))]
    async fn inner_list_commits_by_oid(
        &self,
        request: ListCommitsByOidRequest,
        correlation_id: Option<&MetadataValue<Ascii>>,
    ) -> ResultResponseStream<ListCommitsByOidResponse> {
        tracing_span_id!();
        info!(
            "Processing, request={:?}",
            ListCommitsByOidTracingRequest(&request)
        );

        load_changelog_and_stream(
            self.config.clone(),
            request,
            default_repo_spec_error_status,
            move |req, _repo, cl, tx| {
                for oid_chunk in req.oid.chunks(COMMITS_CHUNK_SIZE) {
                    match commits_by_oid_chunk(cl, oid_chunk) {
                        Ok(resp) => {
                            if !resp.commits.is_empty() {
                                tx.send(Ok(resp));
                            }
                        }
                        Err(e) => tx.send(Err(Status::internal(format!(
                            "Repository corruption {:?}",
                            e
                        )))),
                    }
                }
            },
        )
    }

    #[instrument(
        name = "list_commits_by_ref_name",
        skip(self, request),
        fields(span_id)
    )]
    async fn inner_list_commits_by_ref_name(
        &self,
        request: ListCommitsByRefNameRequest,
        correlation_id: Option<&MetadataValue<Ascii>>,
    ) -> ResultResponseStream<ListCommitsByRefNameResponse> {
        tracing_span_id!();
        info!(
            "Processing, request={:?}",
            ListCommitsByRefNameTracingRequest(&request)
        );

        let config = self.config.clone();
        let store_vfs = repo_store_vfs(&config, &request.repository)
            .await
            .map_err(default_repo_spec_error_status)?;

        load_changelog_and_stream(
            self.config.clone(),
            request,
            default_repo_spec_error_status,
            move |req, _repo, cl, tx| {
                for ref_chunk in req.ref_names.chunks(COMMITS_CHUNK_SIZE) {
                    match commits_by_ref_name_chunk(&store_vfs, cl, ref_chunk) {
                        Ok(resp) => {
                            if !resp.commit_refs.is_empty() {
                                tx.send(Ok(resp));
                            }
                        }
                        Err(e) => tx.send(Err(Status::internal(format!(
                            "Repository corruption {:?}",
                            e
                        )))),
                    }
                }
            },
        )
    }

    #[instrument(name = "find_commits", skip(self, request), fields(span_id))]
    async fn inner_find_commits(
        &self,
        request: &FindCommitsRequest,
        correlation_id: Option<&MetadataValue<Ascii>>,
    ) -> ResultResponseStream<FindCommitsResponse> {
        tracing_span_id!();
        info!(
            "Processing, request={:?}",
            FindCommitsTracingRequest(request)
        );
        if request.limit < 0 {
            return Err(Status::invalid_argument(format!(
                "Got negative limit {}",
                request.limit,
            )));
        }
        let limit = request.limit as usize;
        if request.offset < 0 {
            return Err(Status::invalid_argument(format!(
                "Got negative offset {}",
                request.offset,
            )));
        }
        let offset = request.offset as usize;

        if limit == 0 {
            return empty_response_stream();
        }
        // `FindCommits` gets called a lot with `follow=true` but without `paths` in which case the
        // `follow` has no imlications (must be a default options system on the client side). It is
        // important not to fall back in these cases
        if !request.paths.is_empty() {
            return Err(Status::unimplemented("with paths"));
        }
        if !request.include_referenced_by.is_empty() || request.include_shortstat {
            return Err(Status::unimplemented("with short stats and referenced by"));
        }
        if request.order == find_commits_request::Order::Topo.into() {
            return Err(Status::unimplemented("with topological sorting"));
        }
        if request.all {
            return Err(Status::unimplemented("with 'all' option"));
        }
        if !request.author.is_empty() {
            return Err(Status::unimplemented("with author"));
        }
        if request.skip_merges {
            return Err(Status::unimplemented("with skip_merges option"));
        }
        if request.after.is_some() || request.before.is_some() {
            return Err(Status::unimplemented("with date options"));
        }
        let revision: &[u8] = if request.revision.is_empty() {
            b"HEAD"
        } else {
            &request.revision
        };
        let revision = match GitRevSpec::parse(revision) {
            GitRevSpec::Revision(rev) => rev,
            rs => {
                info!("revision parsed as {:?}", rs);
                return Err(Status::unimplemented("TODO"));
            }
        };

        let config = self.config.clone();
        let store_vfs = repo_store_vfs(&config, &request.repository)
            .await
            .map_err(default_repo_spec_error_status)?;
        match gitlab_revision_node_prefix(&store_vfs, revision)
            .await
            .map_err(|e| Status::internal(format!("Error resolving revision: {:?}", e)))?
        {
            None => {
                // using the sidecar rather than tedious Rust structured errors
                return Err(Status::unimplemented("when revision is not resolved"));
            }
            Some(node_prefix) => {
                info!("Revision resolved as {:x}", &node_prefix);
                load_changelog_and_stream(
                    self.config.clone(),
                    request.clone(),
                    default_repo_spec_error_status,
                    move |_req, _repo, cl, tx| {
                        if let Err(e) =
                            stream_find_commits_ancestors(cl, node_prefix, offset, limit, &tx)
                        {
                            tx.send(Err(Status::internal(format!(
                                "Repository corruption {:?}",
                                e
                            ))));
                        }
                    },
                )
            }
        }
    }

    #[instrument(name = "last_commit_for_path", skip(self, request), fields(span_id))]
    async fn inner_last_commit_for_path(
        &self,
        request: LastCommitForPathRequest,
        correlation_id: Option<&MetadataValue<Ascii>>,
    ) -> Result<Response<LastCommitForPathResponse>, Status> {
        tracing_span_id!();
        info!(
            "Processing, request={:?}",
            LastCommitForPathTracingRequest(&request)
        );
        if request.revision.is_empty() {
            return Err(Status::invalid_argument("empty revision"));
        }

        let config = self.config.clone();
        let store_vfs = repo_store_vfs(&config, &request.repository)
            .await
            .map_err(default_repo_spec_error_status)?;

        let literal_pathspec = request.literal_pathspec
            || if let Some(ref opts) = request.global_options {
                opts.literal_pathspecs
            } else {
                false
            };

        match gitlab_revision_node_prefix(&store_vfs, &request.revision)
            .await
            .map_err(|e| Status::internal(format!("Error resolving revision: {:?}", e)))?
        {
            None => {
                info!("Revision not resolved");
                Ok(Response::new(LastCommitForPathResponse::default()))
            }
            Some(node_prefix) => {
                info!("Revision resolved as {:x}", &node_prefix);

                load_changelog_and_then(
                    self.config.clone(),
                    request,
                    default_repo_spec_error_status,
                    move |req, _repo, cl| {
                        lcfp_in_thread(req, cl, node_prefix, literal_pathspec)
                            .map_err(|e| Status::internal(format!("Repository corruption {:?}", e)))
                    },
                )
                .await
                .map(Response::new)
            }
        }
    }

    #[instrument(name = "find_commit", skip(self, request), fields(span_id))]
    async fn inner_find_commit(
        &self,
        request: FindCommitRequest,
        correlation_id: Option<&MetadataValue<Ascii>>,
    ) -> Result<Response<FindCommitResponse>, Status> {
        tracing_span_id!();
        info!(
            "Processing, request={:?}",
            FindCommitTracingRequest(&request)
        );
        if request.revision.is_empty() {
            return Err(Status::invalid_argument("empty revision"));
        }

        let config = self.config.clone();
        let store_vfs = repo_store_vfs(&config, &request.repository)
            .await
            .map_err(default_repo_spec_error_status)?;

        match gitlab_revision_node_prefix(&store_vfs, &request.revision)
            .await
            .map_err(|e| Status::internal(format!("Error resolving revision: {:?}", e)))?
        {
            None => {
                info!("Revision not resolved");
                Ok(Response::new(FindCommitResponse::default()))
            }
            Some(node_prefix) => {
                info!("Revision resolved as {:x}", &node_prefix);
                let commit = load_changelog_and_then(
                    self.config.clone(),
                    request,
                    default_repo_spec_error_status,
                    move |_req, _repo, cl| {
                        message::commit_for_node_prefix_or_none(cl, node_prefix)
                            .map_err(|e| Status::internal(format!("Repository corruption {:?}", e)))
                    },
                )
                .await?;

                Ok(Response::new(FindCommitResponse { commit }))
            }
        }
    }
}

/// Takes care of boilerplate that would instead be in the startup sequence.
pub fn commit_server(
    config: &Arc<Config>,
    sidecar_servers: &Arc<sidecar::Servers>,
) -> CommitServiceServer<CommitServiceImpl> {
    CommitServiceServer::new(CommitServiceImpl {
        config: config.clone(),
        sidecar_servers: sidecar_servers.clone(),
    })
}
